@@ -38,6 +38,8 @@
  *  Olivier Beyssac <ob@r14.freenix.org>
  */ 
 
+#define _GNU_SOURCE
+
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -45,14 +47,13 @@
 #include <sys/stat.h>
 
 #include <ctype.h>
+#include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <syslog.h>
 #include <signal.h>
 #include <errno.h>
-#include <err.h>
-#include <paths.h>
 #include <stdarg.h>
 #include <pwd.h>
 #include <time.h>
@@ -86,8 +87,8 @@ spthread_t;
 #define CRLF                "\r\n"
 
 #define SMTP_TOOLONG        "500 Line too long" CRLF
-#define SMTP_STARTBUSY      "554 Server Busy" CRLF
-#define SMTP_STARTFAILED    "554 Local Error" CRLF
+#define SMTP_STARTBUSY      "421 Server busy, too many connections" CRLF
+#define SMTP_STARTFAILED    "421 Local Error, cannot start thread" CRLF
 #define SMTP_DATAINTERMED   "354 Start mail input; end with <CRLF>.<CRLF>" CRLF
 #define SMTP_FAILED         "451 Local Error" CRLF
 #define SMTP_NOTSUPP        "502 Command not implemented" CRLF
@@ -111,6 +112,7 @@ spthread_t;
 #define ESMTP_BINARY        "BINARYMIME"
 #define ESMTP_CHECK         "CHECKPOINT"
 #define ESMTP_XCLIENT       "XCLIENT"
+#define ESMTP_XEXCH50       "XEXCH50" 
 
 #define HELO_CMD            "HELO"
 #define EHLO_CMD            "EHLO"
@@ -274,7 +276,7 @@ int sp_run(const char* configfile, const char* pidfile, int dbg_level)
     else if(g_state.outname != NULL && g_state.transparent)
         warnx("the " CFG_OUTADDR " option will be ignored when " CFG_TRANSPARENT " is enabled");
 
-    sp_messagex(NULL, LOG_DEBUG, "starting up...");
+    sp_messagex(NULL, LOG_DEBUG, "starting up (%s)...", VERSION);
 
     /* Drop privileges before daemonizing */
     drop_privileges();
@@ -295,7 +297,16 @@ int sp_run(const char* configfile, const char* pidfile, int dbg_level)
         /* Open the system log */
         openlog(g_state.name, 0, LOG_MAIL);
     }
+    
+    /* Handle some signals */
+    signal(SIGPIPE, SIG_IGN); 
+    signal(SIGHUP, SIG_IGN);
+    signal(SIGINT, on_quit);
+    signal(SIGTERM, on_quit);
 
+    siginterrupt(SIGINT, 1);
+    siginterrupt(SIGTERM, 1);
+    
     /* Create the socket */
     sock = socket(SANY_TYPE(g_state.listenaddr), SOCK_STREAM, 0);
     if(sock < 0)
@@ -316,25 +327,16 @@ int sp_run(const char* configfile, const char* pidfile, int dbg_level)
         sp_message(NULL, LOG_CRIT, "couldn't bind to address: %s", g_state.listenname);
         exit(1);
     }
-
+    
+    sp_messagex(NULL, LOG_DEBUG, "created socket: %s", g_state.listenname);
+    
     /* Let 5 connections queue up */
     if(listen(sock, 5) != 0)
     {
         sp_message(NULL, LOG_CRIT, "couldn't listen on socket");
         exit(1);
     }
-      
-    sp_messagex(NULL, LOG_DEBUG, "created socket: %s", g_state.listenname);
-
-    /* Handle some signals */
-    signal(SIGPIPE, SIG_IGN); 
-    signal(SIGHUP,  SIG_IGN);
-    signal(SIGINT,  on_quit);
-    signal(SIGTERM, on_quit);
-
-    siginterrupt(SIGINT, 1);
-    siginterrupt(SIGTERM, 1);
-
+    
     pid_file(1);
 
     sp_messagex(NULL, LOG_DEBUG, "accepting connections");
@@ -493,7 +495,7 @@ static void connection_loop(int sock)
         /* Set timeouts on client */
         if(setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &(g_state.timeout), sizeof(g_state.timeout)) < 0 ||
            setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &(g_state.timeout), sizeof(g_state.timeout)) < 0)
-            sp_message(NULL, LOG_WARNING, "couldn't set timeouts on incoming connection");        
+            sp_message(NULL, LOG_DEBUG, "couldn't set timeouts on incoming connection");        
 
         fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);    
 
@@ -1048,7 +1050,8 @@ static int smtp_passthru(spctx_t* ctx)
                        is_first_word(p, ESMTP_CHUNK, KL(ESMTP_CHUNK)) ||
                        is_first_word(p, ESMTP_BINARY, KL(ESMTP_BINARY)) ||
                        is_first_word(p, ESMTP_CHECK, KL(ESMTP_CHECK)) ||
-                       is_first_word(p, ESMTP_XCLIENT, KL(ESMTP_XCLIENT)))
+                       is_first_word(p, ESMTP_XCLIENT, KL(ESMTP_XCLIENT)) ||
+                       is_first_word(p, ESMTP_XEXCH50, KL(ESMTP_XEXCH50)))
                     {
                         sp_messagex(ctx, LOG_DEBUG, "filtered ESMTP feature: %s", trim_space((char*)p));
                         
@@ -1134,6 +1137,10 @@ static char* parse_address(char* line)
      * We parse out emails in the form of <blah@blah.com> 
      * as well as accept other addresses.
      */
+    
+    if(strncmp(line, "<>", 2) == 0)
+        return("<>");
+
     if(line[0] == '<')
     {
         if((t = strchr(line, '>')) != NULL)
@@ -1328,7 +1335,6 @@ static void make_date(spctx_t* ctx, char* date)
 {
     size_t date_len;
     struct tm t2;
-    long gmt;
     time_t t;
 
     /* Get a basic date like: 'Wed Jun 30 21:49:08 1993' */    
@@ -1341,14 +1347,50 @@ static void make_date(spctx_t* ctx, char* date)
         return;
     }
       
-    /* Now add the TZ */
     trim_end(date);
-    gmt = t2.tm_gmtoff;
     date_len = strlen(date);
-      
-    snprintf(date + date_len, MAX_DATE_LENGTH - date_len, " %+03d%02d (%s)", 
-             (int)(gmt / 3600), (int)(gmt % 3600), t2.tm_zone);
 
+    {
+#ifdef HAVE_TM_GMTOFF
+        time_t timezone = t2.tm_gmtoff;
+        const char *tzname[2] = { t2.tm_zone, t2.tm_zone };
+
+        snprintf(date + date_len, MAX_DATE_LENGTH - date_len, " %+03d%02d (%s)", 
+                 (int)(timezone / 3600), (int)(timezone % 3600),
+                 tzname[t2.tm_isdst ? 1 : 0]);
+#else
+        /* Apparently Solaris needs this nasty hack.... */
+        #define DAY_MIN         (24 * HOUR_MIN)
+        #define HOUR_MIN        60
+        #define MIN_SEC         60
+
+        struct tm gmt;
+        struct tm *lt;
+        int off;
+
+        gmt = *gmtime(&t);
+        lt = localtime(&t);
+        off = (lt->tm_hour - gmt.tm_hour) * HOUR_MIN + lt->tm_min - gmt.tm_min;
+
+        if (lt->tm_year < gmt.tm_year)
+            off -= DAY_MIN;
+        else if (lt->tm_year > gmt.tm_year)
+            off += DAY_MIN;
+        else if (lt->tm_yday < gmt.tm_yday)
+            off -= DAY_MIN;
+        else if (lt->tm_yday > gmt.tm_yday)
+            off += DAY_MIN;
+        if (lt->tm_sec <= gmt.tm_sec - MIN_SEC)
+            off -= 1;
+        else if (lt->tm_sec >= gmt.tm_sec + MIN_SEC)
+            off += 1;
+
+        snprintf(date + date_len, MAX_DATE_LENGTH - date_len,
+                 " %+03d%02d (%s)", (int)(off / HOUR_MIN), (int)(abs(off) % HOUR_MIN),
+                 tzname[lt->tm_isdst ? 1 : 0]);
+#endif    
+    }
+    
     /* Break it off just in case */
     date[MAX_DATE_LENGTH - 1] = 0;
 }
@@ -1448,14 +1490,21 @@ static int make_header(spctx_t* ctx, const char* format_str, char* header)
 int sp_done_data(spctx_t* ctx)
 {
     FILE* file = 0;
-    int had_header = 0;
     int ret = 0;
-    char line[SP_LINE_LENGTH];
-    char header[MAX_HEADER_LENGTH];
-    size_t header_len;
+    char *line;    
+    char header[MAX_HEADER_LENGTH] = "";
+    size_t header_len, line_len;
+    ssize_t rc;
 
     ASSERT(ctx->cachename[0]);  /* Must still be around */
     ASSERT(!ctx->cachefile);    /* File must be closed */
+    
+    memset(header, 0, sizeof(header));
+
+    /* Alloc line buffer */
+    line_len = SP_LINE_LENGTH;
+    if((line = (char *)malloc(line_len)) == NULL)
+        RETURN(-1);
 
     /* Open the file */
     file = fopen(ctx->cachename, "r");
@@ -1480,7 +1529,7 @@ int sp_done_data(spctx_t* ctx)
             
         sp_messagex(ctx, LOG_DEBUG, "server refused data transfer");
             
-        RETURN(-1);
+        RETURN(0);
     }
 
     sp_messagex(ctx, LOG_DEBUG, "sending from cache file: %s", ctx->cachename);
@@ -1489,16 +1538,16 @@ int sp_done_data(spctx_t* ctx)
         header_len = make_header(ctx, g_state.header, header);
 
     /* If we have to prepend the header, do it */
-    if(header[0] && g_state.header_prepend)
+    if(header[0] != '\0' && g_state.header_prepend)
     {
-	    if(spio_write_data_raw(ctx, &(ctx->server), (char*)header, header_len) == -1 ||
-	       spio_write_data_raw(ctx, &(ctx->server), CRLF, KL(CRLF)) == -1)
+	    if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
+	       spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
 	        RETURN(-1);
-	    had_header = 1;
+		header[0] = '\0';
     }
 
     /* Transfer actual file data */    
-    while(fgets(line, SP_LINE_LENGTH, file) != NULL)
+    while((rc = getline(&line, &line_len, file)) != -1)
     {
         /* 
          * If the line is <CRLF>.<CRLF> we need to change it so that 
@@ -1509,7 +1558,7 @@ int sp_done_data(spctx_t* ctx)
         if(strcmp(line, "." CRLF) == 0)
             strncpy(line, ". " CRLF, SP_LINE_LENGTH);
       
-        if(header[0] && !had_header)
+        if(header[0] != '\0')
         {
             /* 
              * The first blank line we see means the headers are done.
@@ -1517,15 +1566,14 @@ int sp_done_data(spctx_t* ctx)
              */
             if(is_blank_line(line))
             {
-                if(spio_write_data_raw(ctx, &(ctx->server), (char*)header, header_len) == -1 ||
-                   spio_write_data_raw(ctx, &(ctx->server), CRLF, KL(CRLF)) == -1)
+                if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
+                   spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
                     RETURN(-1);
-            
-                had_header = 1;
+                header[0] = '\0';				
             }
         }
         
-        if(spio_write_data_raw(ctx, &(ctx->server), line, strlen(line)) == -1)
+        if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)line, rc) == -1)
             RETURN(-1);
     }
     
@@ -1550,6 +1598,8 @@ int sp_done_data(spctx_t* ctx)
         
 cleanup:
     
+	if(line)
+		free(line); 
     if(file)
         fclose(file); /* read-only so no error check */
     
