@@ -39,14 +39,12 @@
 #include <sys/param.h>
 #include <sys/wait.h>
 
-#include <paths.h>
 #include <ctype.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <syslog.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <err.h>
 #include <signal.h>
 
 #include "usuals.h"
@@ -122,6 +120,10 @@ static int wait_process(spctx_t* sp, pid_t pid, int* status);
  *  STARTUP ETC...
  */
 
+#ifndef HAVE___ARGV
+char** __argv;
+#endif
+
 int main(int argc, char* argv[])
 {
     const char* configfile = DEFAULT_CONFIG;
@@ -131,6 +133,10 @@ int main(int argc, char* argv[])
     int r;
     char* t;
     
+#ifndef HAVE___ARGV
+    __argv = argv;
+#endif
+	
     /* Setup some defaults */
     memset(&g_pxstate, 0, sizeof(g_pxstate));
     g_pxstate.directory = _PATH_TMP;
@@ -293,6 +299,14 @@ void cb_del_context(spctx_t* ctx)
 /* -----------------------------------------------------------------------------
  * IMPLEMENTATION
  */
+
+static void kill_myself()
+{
+    while (1) {
+        kill(getpid(), SIGKILL);
+        sleep(1);
+    }
+}
  
 static pid_t fork_filter(spctx_t* sp, int* infd, int* outfd, int* errfd)
 {
@@ -334,24 +348,27 @@ static pid_t fork_filter(spctx_t* sp, int* infd, int* outfd, int* errfd)
         {
             close(pipe_i[WRITE_END]);
             r = dup2(pipe_i[READ_END], STDIN);
+			close(pipe_i[READ_END]);
         }
         
         if(r >= 0 && outfd)
         {
             close(pipe_o[READ_END]);
             r = dup2(pipe_o[WRITE_END], STDOUT);
+			close(pipe_o[WRITE_END]);
         }
         
         if(r >= 0 && errfd)
         {
             close(pipe_e[READ_END]);
             r = dup2(pipe_e[WRITE_END], STDERR);
+			close(pipe_e[WRITE_END]);
         }
 
         if(r < 0)
         {        
             sp_message(sp, LOG_ERR, "couldn't dup descriptors for filter command");
-            _exit(1);
+            kill_myself();
         }            
 
         /* All the necessary environment vars */        
@@ -362,7 +379,7 @@ static pid_t fork_filter(spctx_t* sp, int* infd, int* outfd, int* errfd)
         
         /* If that returned then there was an error */
         sp_message(sp, LOG_ERR, "error executing the shell for filter command");
-        _exit(1);
+        kill_myself();
         break;        
     };
     
@@ -824,7 +841,7 @@ static void buffer_reject_message(char* data, char* buf, int buflen)
              * Basically if we already have a newline at the end
              * then we need to start a new line 
              */
-            if(buf[strlen(buf)] == '\n')
+			if(buf[strlen(buf) - 1] == '\n')
                 buf[0] = 0;
         }
         else 
